# Convert an academy asset bundle (type 7) to an original one (type 6)
# Steps:
# Find all object pointers, both internal and external, in preload table
# Map internal object pointers (hashed?) to new (indexed) ones wherever they can be found (preload, container, objects)
# Save the asset as format 6

import os
import json
from unitypack.asset import Asset, ObjectPointer, FFOrderedDict, ObjectInfo

APPDATA_PATH = (os.path.expandvars('%userprofile%')) + '/AppData/LocalLow/Unity/Web Player/Cache/FusionFallA/'

srcfile = open(APPDATA_PATH + 'map/CustomAssetBundle-TrainingGrounds', 'rb')

# This particular bundle only has one external reference: library/unity default resources
srcasset = Asset.from_file(srcfile)

pathMap = {}

# The preload table includes every internal object at least once, along with some external objects
for ptr in srcasset.objects[1].contents['m_PreloadTable']:
	# Unity default resources is format 6, so it actually still uses normal incrementing file and path ids
	if ptr.file_id == 1:
		print(ptr.object.asset.name)
	# If we haven't seen this path id before, map it
	elif ptr.path_id not in pathMap:
		# We want the first index to be 2 (index 1 is the AssetBundle object)
		pathMap[ptr.path_id] = len(pathMap) + 2

# We should have found exactly n-1 unique internal objects (again ignoring the AssetBundle at index 1)
assert len(pathMap) == (len(srcasset.objects) - 1)

# Save this path map so it can be used for converting dependent map scenes (this is used in area.py)
with open('info/pathMap.json', 'w') as f: 
	json.dump(pathMap, f, indent=4)

for ptr in srcasset.objects[1].contents['m_PreloadTable']:
	# If it's not in unity default resources, replace the internal pointer with our newly mapped indices
	if ptr.file_id != 1:
		ptr.file_id = 0
		ptr.path_id = pathMap[ptr.path_id]

for pair in srcasset.objects[1].contents['m_Container']:
	ptr = pair[1]['asset']
	# The container will always have only internal assets
	ptr.file_id = 0
	ptr.path_id = pathMap[ptr.path_id]

def replace(obj):
	className = obj.__class__.__name__
	if className == 'GameObject':
		for key, value in obj._obj.items(): obj._obj[key] = replace(value)
		return obj
	elif className == 'ObjectPointer':
		if obj.file_id != 0 and obj.file_id != 1:
			ret = ObjectPointer(None, srcasset)
			ret.file_id = 0
			ret.path_id = pathMap[obj.path_id]
			return ret
		return obj
	# Really long conditional, but I wanted to make sure I knew everything that I was working with
	elif className == 'Transform' or className == 'Animation' or className == 'AnimationClip' or className == 'MeshFilter' or className == 'Mesh' or className == 'Material' or className == 'Texture2D' or className == 'SubMesh' or className == 'Renderer' or className == 'Shader' or className == 'AudioClip' or className == 'EllipsoidParticleEmitter' or className == 'ParticleAnimator' or className == 'ParticleRenderer':
		for key, value in obj._obj.items(): obj._obj[key] = replace(value)
		return obj
	elif className == 'FFOrderedDict':
		for key, value in obj.items(): obj[key] = replace(value)
		return obj
	elif className == 'str' or className == 'int' or className == 'float' or className == 'bool' or className == 'bytes': return obj
	elif className == 'list':
		ret = [0] * len(obj)
		for i in range(len(obj)): ret[i] = replace(obj[i])
		return ret
	elif className == 'tuple':
		ret = [0] * len(obj)
		for i in range(len(obj)): ret[i] = replace(obj[i])
		return tuple(ret)
	elif className == 'NoneType': return obj
	else:
		print(f'Unhandled {className} {obj}')

# Restructure the objects dictionary so the new keys are used
newObjects = {1: srcasset.objects[1]}
for i in srcasset.objects.keys():
	# Ignore asset bundle 
	if i != 1:
		srcasset.objects[i].path_id = pathMap[i]
		newObjects[pathMap[i]] = srcasset.objects[i]
		newObjects[pathMap[i]]._contents = replace(newObjects[pathMap[i]].contents)

# Unity wants these to be ordered, even though the path ids are specified in the object data. Not having the objects strictly put into order like this will crash your game every time you load in.
# So, we restructure the objects dictionary once again.
srcasset._objects = {}
for i in range(1, len(newObjects) + 1):
	srcasset._objects[i] = newObjects[i]

srcasset.format = 6
srcasset.long_object_ids = False

# This assumes the assetbundle directory exists (traininggrounds_2eresourcefile in this case)
APPDATA_PATH = (os.path.expandvars('%userprofile%')) + '/AppData/LocalLow/Unity/Web Player/Cache/FusionFall/'
outfile = open(APPDATA_PATH + 'traininggrounds_2eresourceFile/CustomAssetBundle-TrainingGrounds', 'wb')

srcasset.save(outfile)
outfile.close()